// Scintilla source code edit control
/** @file LexVaraq.cxx
 ** Lexer for the klingon scripting language var'aq
 ** Used for the Scintilla-var'aq-Tutorial
 **/
// Copyright 2013-2014 by Andreas Tscharner <andy@vis.ethz.ch>
// The License.txt file describes the conditions under which this software may be distributed.


#include <cstdlib>
#include <cassert>
#include <cstring>

#include "ILexer.h"
#include "Scintilla.h"
#include "SciLexer.h"

#include "LexAccessor.h"
#include "StyleContext.h"
#include "LexerModule.h"
#include "CharacterSet.h"
#include "WordList.h"


#ifdef SCI_NAMESPACE
using namespace Scintilla;
#endif


static const char *const VaraqWordListDesc[] = {
	"Var\'aq keywords",
	0
};

static const char *const VaraqKeywords = {
	"woD latlh tam chImmoH qaw qawHa\' Hotlh pong cher HIja\'chugh ghobe\'chugh wIv "
	"chov nargh vangqa\' SIj muv ghorqu\' chIm\'a\' tlheghrar naQmoH tlheghrap\'a\' "
	"tlheghpe\' tlheghjuv jor boq boqHa\' boq\'egh boqHa\'\'egh HabboqHa\'\'egh "
	"chuv boqHa\'qa\' loS\'ar wa\'boq wa\'boqHa\' joq joqHa\' qojmI\' qojHa\' "
	"ghurtaH maHghurtaH wejghurtaH poD Hab \'ar mIScher mIS HeHmI\' ghurmI\' "
	"HabmI\'\'a\' mI\'\'a\' mI\'moH mobmoH DuD tlhoch Qo\'moH nIHghoS poSghoS "
	"law\'\'a\' puS\'a\' rap\'a\' law\'rap\'a\' puSrap\'a\' rapbe\'a\' pagh\'a\' "
	"taH\'a\' je joq ghap ghobe\' cha\' \'Ij bep chu\'DonwI\' chu\'tut nuqDaq_jIH "
	"pongmI\' taghDe\'"
};


class LexerVaraq : public ILexer
{
	public:
		LexerVaraq() {}
		virtual ~LexerVaraq() {}

		int SCI_METHOD Version() const {
			return lvOriginal;
		}
		void SCI_METHOD Release() {
			delete this;
		}

		const char * SCI_METHOD PropertyNames() {
			return NULL;
		}
		int SCI_METHOD PropertyType(const char *name) {
			return -1;
		}
		const char * SCI_METHOD DescribeProperty(const char *name) {
			return NULL;
		}
		int SCI_METHOD PropertySet(const char *key, const char *val) {
			return -1;
		}

		const char * SCI_METHOD DescribeWordListSets() {
			return VaraqWordListDesc[0];
		}
		int SCI_METHOD WordListSet(int n, const char *wl) {
			return -1;
		}

		void SCI_METHOD Lex(unsigned int startPos, int lengthDoc, int initStyle, IDocument *pAccess);
		void SCI_METHOD Fold(unsigned int startPos, int lengthDoc, int initStyle, IDocument *pAccess);

		void * SCI_METHOD PrivateCall(int operation, void *pointer) {
			return NULL;
		}

		static ILexer *LexerFactoryVaraq() {
			return new LexerVaraq();
		}
};


void SCI_METHOD LexerVaraq::Lex(unsigned int startPos, int lengthDoc, int initStyle, IDocument *pAccess)
{
	const unsigned int MAX_STR_LEN = 100;


	LexAccessor styler(pAccess);
	StyleContext scCTX(startPos, lengthDoc, initStyle, styler);
	bool pointInNumberUsed = false;
	WordList vqKeywords;
	char *tmpStr;
	CharacterSet setKeywordChars(CharacterSet::setAlpha, "\'");


	vqKeywords.Set(VaraqKeywords);
	tmpStr = new char[MAX_STR_LEN];

	for (; scCTX.More() ; scCTX.Forward()) {
		switch (scCTX.state) {
			case SCE_VARAQ_DEFAULT :
				if (scCTX.Match('(', '*')) {
					scCTX.SetState(SCE_VARAQ_COMMENT);
					scCTX.Forward();
					break;
				};
				if (scCTX.Match('\"')) {
					scCTX.SetState(SCE_VARAQ_STRING);
					break;
				};
				if (IsADigit(scCTX.ch) || (scCTX.Match('-') && IsADigit(scCTX.chNext))) {
					scCTX.SetState(SCE_VARAQ_NUMBER);
					pointInNumberUsed = false;
					break;
				};
				if (setKeywordChars.Contains(scCTX.ch)) {
					scCTX.SetState(SCE_VARAQ_IDENTIFIER);
					break;
				};
				break;

			case SCE_VARAQ_COMMENT :
				if (scCTX.Match('*', ')')) {
					scCTX.Forward();
					scCTX.ForwardSetState(SCE_VARAQ_DEFAULT);
				};
				break;

			case SCE_VARAQ_STRING :
				if (scCTX.Match('\"')) {
					scCTX.ForwardSetState(SCE_VARAQ_DEFAULT);
				};
				break;

			case SCE_VARAQ_NUMBER :
				if (scCTX.Match('.')) {
					if (pointInNumberUsed) {
						scCTX.SetState(SCE_VARAQ_DEFAULT);
					} else {
						pointInNumberUsed = true;
					};
				} else {
					if (!IsADigit(scCTX.ch)) {
						scCTX.SetState(SCE_VARAQ_DEFAULT);
					};
				};
				break;

			case SCE_VARAQ_IDENTIFIER :
				if (!setKeywordChars.Contains(scCTX.ch)) {
					memset(tmpStr, 0, sizeof(char)*MAX_STR_LEN);
					scCTX.GetCurrent(tmpStr, MAX_STR_LEN);

					if (vqKeywords.InList(tmpStr)) {
						scCTX.ChangeState(SCE_VARAQ_KEYWORD);
					};
					scCTX.SetState(SCE_VARAQ_DEFAULT);
				};
				break;
		};
	};
	scCTX.Complete();

	delete[] tmpStr;
}

void SCI_METHOD LexerVaraq::Fold(unsigned int startPos, int lengthDoc, int initStyle, IDocument *pAccess)
{
	LexAccessor styler(pAccess);
	unsigned int endPos = startPos + lengthDoc;
	char chNext = styler[startPos];
	int lineCurrent = styler.GetLine(startPos);
	int levelPrev = styler.LevelAt(lineCurrent) & SC_FOLDLEVELNUMBERMASK;
	int levelCurrent = levelPrev;
	char ch;
	bool atEOL;


	for (unsigned int i = startPos; i < endPos; i++) {
		ch = chNext;
		chNext = styler.SafeGetCharAt(i+1);

		atEOL = ((ch == '\r' && chNext != '\n') || (ch == '\n'));

		if (ch == '{') {
			levelCurrent++;
		};
		if (ch == '}') {
			levelCurrent--;
		};

		if (atEOL || (i == (endPos-1))) {
			int lev = levelPrev;

			if (levelCurrent > levelPrev) {
				lev |= SC_FOLDLEVELHEADERFLAG;
			};

			if (lev != styler.LevelAt(lineCurrent)) {
				styler.SetLevel(lineCurrent, lev);
			};
			lineCurrent++;
			levelPrev = levelCurrent;
		};
	};
}


LexerModule lmVaraq(SCLEX_VARAQ, LexerVaraq::LexerFactoryVaraq, "varaq", VaraqWordListDesc);
